##
# File:    RepositoryProvider.py
# Author:  J. Westbrook
# Date:    21-Mar-2018
#
# Updates:
#   22-Mar-2018  jdw add support for all repositories -
#   26-Mar-2018  jdw internalize the use of externally provided configuration object -
#   27-Mar-2018  jdw add path to support mock repositories for testing.
#   23-May-2018  jdw add getRepoPathList() convenience method
#   18-Jun-2018  jdw move mock support to the configuration module
#   12-Jul-2018  jdw correct config for PDBX_REPO_PATH
#   13-Aug-2018  jdw add support for gz compressed entry files
#   24-Oct-2018  jdw update for new configuration organization
#   28-Nov-2018  jdw add mergeBirdRefData()
#   13-Dec-2018  jdw add preliminary I/HM repository support
#    5-Feb-2019  jdw add just method naming conventions, add getLocator() method,
#                    consolidate deliver of path configuration details in __getRepoTopPath().
#   14-Mar-2019  jdw add VRPT_REPO_PATH_ENV as an override for the validation report repo path.
#   27-Aug-2019  jdw filter missing validation reports
#   16-Sep-2019  jdw consolidate chem_comp_core with bird_chem_comp_core
#   14-Feb-2020  jdw migrate to rcsb.utils.repository
#   17-Sep-2021  jdw add remote repository access methods
#   29-Sep-2021  jdw make default discoveryMode a configuration option add inputIdCodeList to getLocatorObjList()
#    8-Oct-2021  jdw pass configuration URLs to CurrentHoldingsProvider and RemoveHoldingsProvider
#                    ValidationReportProvider() migrated to ValidationReportAdapter()
#    8-Oct-2021  jdw add warning messages for empty read/merge container results in method __mergeContainers()
#    5-Apr-2022  dwp Add support for loading id code lists for bird_chem_comp_core (mainly used for Azure testing)
#   13-Apr-2022  dwp Update methods for obtaining list of computed-model files
#    3-Aug-2022  dwp Enable retrieval of specific computed-model files with input
#    2-Feb-2023  dwp add support for requesting specific inputIdCodeList/idCodeList for CSMs
#   12-Jun-2023  dwp disable useCache for holdings files to force fresh re-download
#    5-Mar-2024  dwp Adjustments to support BCIF loading and CSM scaling (eliminate need to import CSM holdings file)
#   19-Mar-2024  dwp Raise exception and return empty list if not all dataContainers are properly read from file
#                    in __mergeContainers() (e.g., if mmCIF file is read in but validation report fails)
##
"""
Utilities for scanning and accessing data in PDBx/mmCIF data in common repository file systems or via remote repository services.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.io.HashableDict import HashableDict
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil
from rcsb.utils.repository.CurrentHoldingsProvider import CurrentHoldingsProvider
from rcsb.utils.repository.RemovedHoldingsProvider import RemovedHoldingsProvider
from rcsb.utils.validation.ValidationReportAdapter import ValidationReportAdapter

logger = logging.getLogger(__name__)


def toCifWrapper(xrt):
    dirPath = os.environ.get("_RP_DICT_PATH_")
    vpr = ValidationReportAdapter(dirPath=dirPath, useCache=True)
    vrd = vpr.getReader()
    return vrd.toCif(xrt)


class RepositoryProvider(object):
    """Utilities for scanning and accessing data in PDBx/mmCIF data in common repository file systems or via remote repository services.

    Discovery modes:

    Local file system mode - data sets are discovered by scanning repository files systems for files matching the require patterns,
                     and repository paths are defined in configuration data. File system path information may be provided
                     explicitly to avoid file system scans. Some merging and consolidation features are provided for validation data
                     and chemical reference data.

    Remote access mode - data sets are discovered from reading repository inventory files,
                     and repository service endpoints are defined in configuration data.


    """

    def __init__(self, cfgOb, cachePath=None, numProc=8, fileLimit=None, verbose=False, discoveryMode=None):
        self.__fileLimit = fileLimit
        self.__numProc = numProc
        self.__verbose = verbose
        self.__cfgOb = cfgOb
        self.__configName = self.__cfgOb.getDefaultSectionName()
        #
        self.__discoveryMode = discoveryMode if discoveryMode else self.__cfgOb.get("DISCOVERY_MODE", sectionName=self.__configName, default="local")
        self.__baseUrlPDB = self.__cfgOb.getPath("PDB_REPO_URL", sectionName=self.__configName, default="https://files.wwpdb.org/pub")
        self.__fallbackUrlPDB = self.__cfgOb.getPath("PDB_REPO_FALLBACK_URL", sectionName=self.__configName, default="https://files.wwpdb.org/pub")
        self.__baseUrlPDBDev = self.__cfgOb.getPath("PDBDEV_REPO_URL", sectionName=self.__configName, default="https://pdb-dev.wwpdb.org")
        self.__edMapUrl = self.__cfgOb.getPath("RCSB_EDMAP_LIST_PATH", sectionName=self.__configName, default=None)
        #
        self.__kwD = {
            "holdingsTargetUrl": os.path.join(self.__baseUrlPDB, "pdb", "holdings"),
            "holdingsFallbackUrl": os.path.join(self.__fallbackUrlPDB, "pdb", "holdings"),
            "edmapsLocator": self.__edMapUrl,
            "updateTargetUrl": os.path.join(self.__baseUrlPDB, "pdb", "data", "status", "latest"),
            "updateFallbackUrl": os.path.join(self.__fallbackUrlPDB, "pdb", "data", "status", "latest"),
            "filterType": "assign-dates",
        }
        #
        self.__topCachePath = cachePath if cachePath else "."
        self.__cachePath = os.path.join(self.__topCachePath, "repo_util")
        #
        self.__mU = MarshalUtil(workPath=self.__cachePath)
        self.__fU = FileUtil(workPath=self.__cachePath)
        self.__chP = None
        self.__rhP = None
        logger.info("Discovery mode is %r", self.__discoveryMode)
        #

    def getLocatorObjList(self, contentType, inputPathList=None, inputIdCodeList=None, mergeContentTypes=None, excludeIds=None):
        """Convenience method to get the data path list for the input repository content type.

        Args:
            contentType (str): Repository content type (e.g. pdbx, chem_comp, bird, ...)
            inputPathList (list, optional): path list that will be returned if provided (discoveryMode=local).
            inputIdCodeList (list, optional): locators will be returned for this ID code list (discoveryMode=remote).
            mergeContentTypes (list, optional): repository content types to combined with the
                                primary content type.
            excludeIds (list or dict): exclude any locators for idCodes in this list or dictionary

        Returns:
            (list): simple list of data file paths OR a tuple containing file path, format and merge details


        Example:

            locator object ({"locator": <path or URI> "fmt": <supported format code>, 'kwargs': {}},{}, ... )

             supported extensions are:
                 kwargs : {"marshalHelper": <fuction applied post read (e.g. toCifWrapper)>}

             multiple artifacts within the a locator may be optionally merged/consolidated into a single data object.

        """
        inputPathList = inputPathList if inputPathList else []
        inputIdCodeList = inputIdCodeList if inputIdCodeList else []
        if inputPathList:
            return self.__getLocatorObjListWithInput(contentType, inputPathList=inputPathList, mergeContentTypes=mergeContentTypes)
        #
        locatorList = self.__getLocatorList(contentType, inputPathList=inputPathList, inputIdCodeList=inputIdCodeList, mergeContentTypes=mergeContentTypes)
        #
        if excludeIds:
            fL = []
            for locator in locatorList:
                if isinstance(locator, str):
                    pth = locator
                else:
                    pth = locator[0]["locator"]
                #
                idCode = self.__getIdcodeFromLocatorPath(contentType, pth)
                if idCode in excludeIds:
                    continue
                fL.append(locator)
            locatorList = fL

        return locatorList

    def getContainerList(self, locatorObjList):
        """Return the PDBx data container list obtained by parsing the input locator object list."""
        cL = []
        for locatorObj in locatorObjList:
            myContainerList = self.__mergeContainers(locatorObj, fmt="mmcif", mergeTarget=0)
            for cA in myContainerList:
                cL.append(cA)
        return cL

    def getLocatorIdcodes(self, contentType, locatorObjList, locatorIndex=0):
        try:
            if locatorObjList and isinstance(locatorObjList[0], str):
                return [self.__getIdcodeFromLocatorPath(contentType, pth) for pth in locatorObjList]
            else:
                return [self.__getIdcodeFromLocatorPath(contentType, locatorObj[locatorIndex]["locator"]) for locatorObj in locatorObjList]
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return []

    def getLocatorPaths(self, locatorObjList, locatorIndex=0):
        try:
            if locatorObjList and isinstance(locatorObjList[0], str):
                return locatorObjList
            else:
                return [locatorObj[locatorIndex]["locator"] for locatorObj in locatorObjList]
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return []

    def getLocatorsFromPaths(self, locatorObjList, pathList, locatorIndex=0):
        """Return locator objects with paths (locatorObjIndex) matching the input pathList."""
        # index the input locatorObjList
        rL = []
        try:
            if locatorObjList and isinstance(locatorObjList[0], str):
                return pathList
            #
            locIdx = {}
            for ii, locatorObj in enumerate(locatorObjList):
                if "locator" in locatorObj[locatorIndex]:
                    locIdx[locatorObj[locatorIndex]["locator"]] = ii
            #
            for pth in pathList:
                jj = locIdx[pth] if pth in locIdx else None
                if jj is not None:
                    rL.append(locatorObjList[jj])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return rL

    #  ---- Private methods ----
    def __getLocatorObjListWithInput(self, contentType, inputPathList=None, mergeContentTypes=None):
        """Convenience method to get the data path list for the input repository content type.
        This is a special case to handle the content merging for the input path/locator list.

        Args:
            contentType (str): Repository content type (e.g. pdbx, chem_comp, bird, ...)
            inputPathList (list, optional): path list that will be returned if provided.
            mergeContentTypes (list, optional): repository content types to combined with the
                                primary content type.

        Returns:
            Obj list: data file paths or tuple of file paths

        """
        inputPathList = inputPathList if inputPathList else []
        locatorList = self.__getLocatorList(contentType, inputPathList=inputPathList)

        if mergeContentTypes and "vrpt" in mergeContentTypes and contentType in ["pdbx", "pdbx_core"]:
            dictPath = os.path.join(self.__topCachePath, self.__cfgOb.get("DICTIONARY_CACHE_DIR", sectionName=self.__cfgOb.getDefaultSectionName()))
            os.environ["_RP_DICT_PATH_"] = dictPath
            #
            locObjL = []
            for locator in locatorList:
                if isinstance(locator, str):
                    kwD = HashableDict({})
                    oL = [HashableDict({"locator": locator, "fmt": "mmcif", "kwargs": kwD})]
                    for mergeContentType in mergeContentTypes:
                        _, fn = os.path.split(locator)
                        idCode = fn[:4] if fn and len(fn) >= 8 else None
                        mergeLocator = self.__getLocator(mergeContentType, idCode, checkExists=True) if idCode else None
                        if mergeLocator:
                            # kwD = HashableDict({"marshalHelper": vrd.toCif})
                            kwD = HashableDict({"marshalHelper": toCifWrapper})
                            oL.append(HashableDict({"locator": mergeLocator, "fmt": "xml", "kwargs": kwD}))
                    lObj = tuple(oL)
                else:
                    logger.error("Unexpected output locator type %r", locator)
                    lObj = locator
                locObjL.append(lObj)
            #
            locatorList = locObjL
        # -
        if contentType in ["pdbx_comp_model_core"]:
            locObjL = []
            for inputPath in inputPathList:
                if isinstance(inputPath, str):
                    if inputPath.strip() in locatorList:
                        locObjL.append(inputPath.strip())
            #
            locatorList = locObjL
        # -
        return locatorList

    def __mergeContainers(self, locatorObj, fmt="mmcif", mergeTarget=0):
        """Consolidate content in auxiliary files locatorObj[1:] into the locatorObj[0] container index 'mergeTarget'."""
        #
        cL = []
        try:
            if isinstance(locatorObj, str):
                # This is followed for CSMs and anything else that doesn't have an associated validation report file
                if locatorObj.lower().endswith(".bcif.gz") or locatorObj.lower().endswith(".bcif"):
                    fmt = "bcif"
                logger.debug("locatorObj %s fmt %s", locatorObj, fmt)
                cL = self.__mU.doImport(locatorObj, fmt=fmt)
                if not cL:
                    logger.error("locator %r fmt %s returned empty container list.", locatorObj, fmt)
                    raise ValueError("locator %r fmt %s returned empty container list" % (locatorObj, fmt))
            #
            elif isinstance(locatorObj, (list, tuple)) and locatorObj:
                # This is followed for Experimental mmCIF files (anything with have an associated validation report file)
                dD = locatorObj[0]
                kw = dD["kwargs"]
                cL = self.__mU.doImport(dD["locator"], fmt=dD["fmt"], **kw)
                if cL:
                    for dD in locatorObj[1:]:
                        kw = dD["kwargs"]
                        rObj = self.__mU.doImport(dD["locator"], fmt=dD["fmt"], **kw)
                        mergeL = rObj if rObj else []
                        if not mergeL:
                            logger.error("locator object with leading path %r returned empty container list (%r)", dD["locator"], locatorObj)
                            raise ValueError("locator object with leading path %r returned empty container list (%r)" % (dD["locator"], locatorObj))
                        for mc in mergeL:
                            cL[mergeTarget].merge(mc)
                else:
                    logger.error("locator object with leading path %r returned empty container list (%r)", dD["locator"], locatorObj)
                    raise ValueError("locator object with leading path %r returned empty container list (%r)" % (dD["locator"], locatorObj))
            #
            else:
                logger.warning("non-comforming locator object %r", locatorObj)
        #
        except Exception as e:
            logger.exception("Failing for %r with %s", locatorObj, str(e))
            return []
        #
        return cL if cL else []

    def __getLocatorList(self, contentType, inputPathList=None, inputIdCodeList=None, mergeContentTypes=None):
        if self.__discoveryMode == "local":
            return self.__getLocatorListLocal(contentType, inputPathList=inputPathList, mergeContentTypes=mergeContentTypes)
        else:
            return self.__getLocatorListRemote(contentType, inputIdCodeList=inputIdCodeList, mergeContentTypes=mergeContentTypes)

    def __getLocatorListLocal(self, contentType, inputPathList=None, mergeContentTypes=None):
        """Internal convenience method to return repository local path lists by content type:"""
        outputLocatorList = []
        inputPathList = inputPathList if inputPathList else []
        try:
            if contentType in ["bird", "bird_core"]:
                outputLocatorList = inputPathList if inputPathList else self.__getBirdPathList()
            elif contentType == "bird_family":
                outputLocatorList = inputPathList if inputPathList else self.__getBirdFamilyPathList()
            elif contentType in ["chem_comp"]:
                outputLocatorList = inputPathList if inputPathList else self.__getChemCompPathList()
            elif contentType in ["bird_chem_comp"]:
                outputLocatorList = inputPathList if inputPathList else self.__getBirdChemCompPathList()
            elif contentType in ["pdbx", "pdbx_core"] and mergeContentTypes and "vrpt" in mergeContentTypes:
                dictPath = os.path.join(self.__topCachePath, self.__cfgOb.get("DICTIONARY_CACHE_DIR", sectionName=self.__cfgOb.getDefaultSectionName()))
                os.environ["_RP_DICT_PATH_"] = dictPath
                outputLocatorList = self.__getEntryLocatorObjList(mergeContentTypes=mergeContentTypes)

            elif contentType in ["pdbx", "pdbx_core"]:
                outputLocatorList = inputPathList if inputPathList else self.__getEntryPathList()
            #
            elif contentType in ["pdbx_obsolete"]:
                outputLocatorList = inputPathList if inputPathList else self.getObsoleteEntryPathList()
            elif contentType in ["chem_comp_core", "bird_consolidated", "bird_chem_comp_core"]:
                outputLocatorList = inputPathList if inputPathList else self.mergeBirdAndChemCompRefData()
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:
                outputLocatorList = inputPathList if inputPathList else self.__getIhmDevPathList()
            elif contentType in ["pdb_distro", "da_internal", "status_history"]:
                outputLocatorList = inputPathList if inputPathList else []
            elif contentType in ["pdbx_comp_model_core"]:
                outputLocatorList = inputPathList if inputPathList else self.__getCompModelPathList()
            else:
                logger.warning("Unsupported contentType %s", contentType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        if self.__fileLimit:
            outputLocatorList = outputLocatorList[: self.__fileLimit]

        return sorted(outputLocatorList) if outputLocatorList and isinstance(outputLocatorList[0], str) else outputLocatorList

    def __getLocatorListRemote(self, contentType, inputIdCodeList=None, mergeContentTypes=None):
        outputLocatorList = []
        idCodeList = inputIdCodeList if inputIdCodeList else []
        logger.info("Getting remote locator list for contentType %s with idCodeList length (%d)", contentType, len(idCodeList))
        try:
            if contentType in ["bird", "bird_core"]:
                outputLocatorList = self.__getBirdUriList(idCodeList=idCodeList)
            elif contentType == "bird_family":
                outputLocatorList = self.__getBirdFamilyUriList(idCodeList=idCodeList)
            elif contentType in ["chem_comp"]:
                outputLocatorList = self.__getChemCompUriList(idCodeList=idCodeList)
            elif contentType in ["bird_chem_comp"]:
                outputLocatorList = self.__getBirdChemCompUriList(idCodeList=idCodeList)
            elif contentType in ["chem_comp_core", "bird_consolidated", "bird_chem_comp_core"] and not self.__fileLimit:
                outputLocatorList = self.mergeBirdAndChemCompRefData()
            elif contentType in ["chem_comp_core", "bird_consolidated", "bird_chem_comp_core"] and self.__fileLimit:
                outputLocatorList = self.mergeBirdAndChemCompRefDataWithInput(idCodeList=idCodeList)
            #
            elif contentType in ["pdbx", "pdbx_core"]:
                if mergeContentTypes and "vrpt" in mergeContentTypes:
                    dictPath = os.path.join(self.__topCachePath, self.__cfgOb.get("DICTIONARY_CACHE_DIR", sectionName=self.__cfgOb.getDefaultSectionName()))
                    os.environ["_RP_DICT_PATH_"] = dictPath
                    outputLocatorList = self.__getEntryUriList(idCodeList=idCodeList, mergeContentTypes=mergeContentTypes)
                else:
                    outputLocatorList = self.__getEntryUriList(idCodeList=idCodeList)
            #
            elif contentType in ["pdbx_obsolete"]:
                outputLocatorList = self.__getObsoleteEntryUriList(idCodeList=idCodeList)
            #
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:
                outputLocatorList = self.__getIhmDevPathList()
            elif contentType in ["pdbx_comp_model_core"]:
                outputLocatorList = self.__getCompModelPathList(idCodeList=idCodeList, fmt="mmcif")
                # outputLocatorList = self.__getCompModelPathList(idCodeList=idCodeList, fmt="bcif")
            else:
                logger.warning("Unsupported contentType %s", contentType)

        except Exception as e:
            logger.exception("Failing with %s", str(e))

        logger.debug("outputLocatorList before applying fileLimit (%r): %r", self.__fileLimit, outputLocatorList)
        if self.__fileLimit:
            outputLocatorList = outputLocatorList[: self.__fileLimit]
            logger.debug("outputLocatorList after applying fileLimit (%r): %r", self.__fileLimit, outputLocatorList)

        return sorted(outputLocatorList) if outputLocatorList and isinstance(outputLocatorList[0], str) else outputLocatorList

    def __getLocator(self, contentType, idCode, version="v1-0", checkExists=False):
        if self.__discoveryMode == "local":
            return self.__getLocatorLocal(contentType, idCode, version=version, checkExists=checkExists)
        else:
            return self.__getLocatorRemote(contentType, idCode)

    def __getLocatorLocal(self, contentType, idCode, version="v1-0", checkExists=False):
        """Convenience method to return repository path for a content type and cardinal identifier."""
        pth = None
        try:
            idCodel = idCode.lower()
            if contentType == "bird":
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode[-1], idCode + ".cif")
            elif contentType == "bird_family":
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode[-1], idCode + ".cif")
            elif contentType in ["chem_comp", "chem_comp_core"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode[0], idCode, idCode + ".cif")
            elif contentType in ["bird_chem_comp"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode[-1], idCode + ".cif")
            elif contentType in ["pdbx", "pdbx_core", "pdbx_obsolete"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCodel[1:3], idCodel + ".cif.gz")
            elif contentType in ["bird_consolidated", "bird_chem_comp_core"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode + ".cif")
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCode, idCode + "_model_%s.cif.gz" % version)
            elif contentType in ["pdb_distro", "da_internal", "status_history"]:
                pass
            elif contentType in ["vrpt"]:
                pth = os.path.join(self.__getRepoLocalPath(contentType), idCodel[1:3], idCodel, idCodel + "_validation.xml.gz")
            else:
                logger.warning("Unsupported local contentType %s", contentType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        if checkExists:
            pth = pth if self.__mU.exists(pth) else None
        return pth

    def __getLocatorRemote(self, contentType, idCode, repositoryLayout="pdbftp"):
        """Convenience method to return the URI for a content type and cardinal identifier."""
        uri = None
        _ = repositoryLayout
        try:
            idCodel = idCode.lower()
            if contentType == "bird":
                # /pdb/refdata/bird/prd/1/
                uri = os.path.join(self.__baseUrlPDB, "pdb", "refdata", "bird", "prd", idCode[-1], idCode + ".cif")
            elif contentType == "bird_family":
                uri = os.path.join(self.__baseUrlPDB, "pdb", "refdata", "bird", "family", idCode[-1], idCode + ".cif")
            elif contentType in ["bird_chem_comp"]:
                uri = os.path.join(self.__baseUrlPDB, "pdb", "refdata", "bird", "prdcc", idCode[-1], idCode + ".cif")
            elif contentType in ["chem_comp", "chem_comp_core"]:
                uri = os.path.join(self.__baseUrlPDB, "pdb", "refdata", "chem_comp", idCode[-1], idCode, idCode + ".cif")
            #
            elif contentType in ["pdbx", "pdbx_core"]:
                # pdb/data/structures/divided/mmCIF
                uri = os.path.join(self.__baseUrlPDB, "pdb", "data", "structures", "divided", "mmCIF", idCodel[1:3], idCodel + ".cif.gz")
            elif contentType in ["vrpt", "validation_report"]:
                # /pdb/validation_reports/
                # https://files.wwpdb.org/pub/pdb/validation_reports/00/100d/100d_validation.xml.gz
                uri = os.path.join(self.__baseUrlPDB, "pdb", "validation_reports", idCodel[1:3], idCodel, idCodel + "_validation.xml.gz")
                # logger.info("uri %r", uri)
            #
            elif contentType in ["pdbx_obsolete"]:
                # pdb/data/structures/obsolete/mmCIF/
                uri = os.path.join(self.__baseUrlPDB, "pdb", "data", "structures", "obsolete", "mmCIF", idCodel[1:3], idCodel + ".cif.gz")
            elif contentType in ["bird_consolidated", "bird_chem_comp_core"]:
                uri = os.path.join(self.__getRepoLocalPath(contentType), idCode + ".cif")
            #
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:

                # https://pdb-dev.wwpdb.org/cif/PDBDEV_00000001.cif
                uri = os.path.join(self.__baseUrlPDBDev, "cif", idCode + ".cif")
            elif contentType in ["pdb_distro", "da_internal", "status_history"]:
                pass
            else:
                logger.warning("Unsupported remote contentType %s", contentType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return uri

    def __getIdcodeFromLocatorPath(self, contentType, pth):
        """Convenience method to return the idcode from the locator path."""
        idCode = None
        try:
            bn = os.path.basename(pth)
            if contentType in ["pdbx", "pdbx_core", "pdbx_obsolete", "bird", "bird_family", "chem_comp", "chem_comp_core", "bird_consolidated", "bird_chem_comp_core"]:
                idCode = bn.split(".")[0]
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:
                tC = bn.split(".")[0]
                idCode = "_".join(tC.split("_")[:2])
            elif contentType in ["pdb_distro", "da_internal", "status_history"]:
                idCode = None
            elif contentType in ["vrpt"]:
                tC = bn.split(".")[0]
                idCode = tC.split("_")[0]
            else:
                logger.warning("Unsupported contentType %s", contentType)
            idCode = idCode.upper() if idCode else None
        except Exception as e:
            logger.exception("Failing for %r %r with %s", contentType, pth, str(e))
        return idCode

    def __getRepoLocalPath(self, contentType):
        """Convenience method to return repository top path from configuration data."""
        pth = None
        try:
            if contentType == "bird":
                pth = self.__cfgOb.getPath("BIRD_REPO_PATH", sectionName=self.__configName)
            elif contentType == "bird_family":
                pth = self.__cfgOb.getPath("BIRD_FAMILY_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["chem_comp", "chem_comp_core"]:
                pth = self.__cfgOb.getPath("CHEM_COMP_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["bird_chem_comp"]:
                pth = self.__cfgOb.getPath("BIRD_CHEM_COMP_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["pdbx", "pdbx_core"]:
                pth = self.__cfgOb.getPath("PDBX_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["pdbx_obsolete"]:
                pth = self.__cfgOb.getPath("PDBX_OBSOLETE_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["pdbx_comp_model_core"]:
                pth = self.__cfgOb.getPath("PDBX_COMP_MODEL_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["bird_consolidated", "bird_chem_comp_core"]:
                pth = self.__cachePath
            elif contentType in ["ihm_dev", "ihm_dev_core", "ihm_dev_full"]:
                pth = self.__cfgOb.getPath("IHM_DEV_REPO_PATH", sectionName=self.__configName)
            elif contentType in ["pdb_distro", "da_internal", "status_history"]:
                pass
            elif contentType in ["vrpt"]:
                pth = self.__cfgOb.getEnvValue("VRPT_REPO_PATH_ENV", sectionName=self.__configName, default=None)
                if pth is None:
                    pth = self.__cfgOb.getPath("VRPT_REPO_PATH", sectionName=self.__configName)
                else:
                    logger.debug("Using validation report path from environment assignment %s", pth)
            else:
                logger.warning("Unsupported contentType %s", contentType)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return pth

    # JDW ---  URI code ----
    def __getEntryUriList(self, idCodeList=None, mergeContentTypes=None):
        uL = []
        try:
            if not self.__chP:
                logger.info("Initializing CurrentHoldingsProvider")
                self.__chP = CurrentHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            ok = self.__chP.testCache()
            logger.info("CurrentHoldingsProvider testCache (%r)", ok)
            #
            tIdL = self.__chP.getEntryIdList()
            logger.info("original tIdL length (%r)", len(tIdL))
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                # idCodeList = [t.upper() for t in idCodeList]
                # tIdL = list(set(tIdL).intersection(idCodeList))
                logger.debug("idCodeList selected tIdL: %r", tIdL)
                logger.info("idCodeList selected tIdL length (%r)", len(tIdL))
                if len(tIdL) > 10:
                    logger.info("idCodeList selected tIdL first few: %r", tIdL[0:5])
            #
            if not (mergeContentTypes and "vrpt" in mergeContentTypes):
                logger.error("Validation mergeContentTypes not enabled!")
            #
            for tId in tIdL:
                kwD = HashableDict({})
                locObj = [HashableDict({"locator": self.__getLocatorRemote("pdbx_core", tId), "fmt": "mmcif", "kwargs": kwD})]
                if mergeContentTypes and "vrpt" in mergeContentTypes:
                    # if self.__chP.hasEntryContentType(tId, "validation_report"):
                    if self.__chP.hasValidationReportData(tId):
                        kwD = HashableDict({"marshalHelper": toCifWrapper})
                        locObj.append(HashableDict({"locator": self.__getLocatorRemote("validation_report", tId), "fmt": "xml", "kwargs": kwD}))
                    else:
                        logger.warning("Validation data not found for id %r", tId)
                uL.append(tuple(locObj))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    def __getObsoleteEntryUriList(self, idCodeList=None):
        uL = []
        try:
            if not self.__rhP:
                self.__rhP = RemovedHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            tIdL = self.__rhP.getEntryByStatus("OBS")
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                logger.info("idCodeList selected: %r", tIdL)
            #
            for tId in tIdL:
                kwD = HashableDict({})
                locObj = [HashableDict({"locator": self.__getLocatorRemote("pdbx_obsolete", tId), "fmt": "mmcif", "kwargs": kwD})]
                uL.append(tuple(locObj))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    def __getBirdUriList(self, idCodeList=None):
        uL = []
        try:
            if not self.__chP:
                self.__chP = CurrentHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            tIdL = self.__chP.getBirdIdList()
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                logger.info("idCodeList selected: %r", tIdL)
            #
            kwD = HashableDict({})
            for tId in tIdL:
                uL.append(tuple([HashableDict({"locator": self.__getLocatorRemote("bird", tId), "fmt": "mmcif", "kwargs": kwD})]))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    def __getBirdFamilyUriList(self, idCodeList=None):
        uL = []
        try:
            if not self.__chP:
                self.__chP = CurrentHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            tIdL = self.__chP.getBirdFamilyIdList()
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                logger.info("idCodeList selected: %r", tIdL)
            #
            kwD = HashableDict({})
            for tId in tIdL:
                uL.append(tuple([{"locator": self.__getLocatorRemote("bird_family", tId), "fmt": "mmcif", "kwargs": kwD}]))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    def __getChemCompUriList(self, idCodeList=None):
        uL = []
        try:
            if not self.__chP:
                self.__chP = CurrentHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            tIdL = self.__chP.getChemCompIdList()
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                logger.info("idCodeList selected: %r", tIdL)
            #
            kwD = HashableDict({})
            for tId in tIdL:
                uL.append(tuple([HashableDict({"locator": self.__getLocatorRemote("chem_comp", tId), "fmt": "mmcif", "kwargs": kwD})]))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    def __getBirdChemCompUriList(self, idCodeList=None):
        uL = []
        try:
            if not self.__chP:
                self.__chP = CurrentHoldingsProvider(self.__topCachePath, **self.__kwD)
            #
            tIdL = self.__chP.getBirdChemCompIdList()
            if idCodeList:
                tIdD = dict.fromkeys(tIdL, True)
                tIdL = [idCode.upper() for idCode in idCodeList if idCode.upper() in tIdD]
                logger.info("idCodeList selected: %r", tIdL)
            #
            kwD = HashableDict({})
            for tId in tIdL:
                uL.append(tuple([HashableDict({"locator": self.__getLocatorRemote("bird_chem_comp", tId), "fmt": "mmcif", "kwargs": kwD})]))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(uL)

    # -- Path based code ---
    def _chemCompPathWorker(self, dataList, procName, optionsD, workingDir):
        """Return the list of chemical component definition file paths in the current repository."""
        _ = procName
        _ = workingDir
        topRepoPath = optionsD["topRepoPath"]
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, _, files in os.walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.endswith(".cif") and len(name) <= 7:
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def __getChemCompPathList(self):
        return self.__fetchChemCompPathList(self.__getRepoLocalPath("chem_comp"), numProc=self.__numProc)

    def __fetchChemCompPathList(self, topRepoPath, numProc=8):
        """Get the path list for the chemical component definition repository"""
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s", ts)
        startTime = time.time()
        pathList = []
        try:
            dataS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
            dataList = [a for a in dataS]
            optD = {}
            optD["topRepoPath"] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_chemCompPathWorker")
            _, _, retLists, _ = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds", len(pathList), endTime0 - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(pathList)

    def _entryLocatorObjWithMergeWorker(self, dataList, procName, optionsD, workingDir):
        """Return the list of entry locator objects including merge content in the current repository."""
        _ = procName
        _ = workingDir
        topRepoPath = optionsD["topRepoPath"]
        mergeContentTypes = optionsD["mergeContentTypes"]
        locatorObjList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, _, files in os.walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for fn in files:
                    if (fn.endswith(".cif.gz") and len(fn) == 11) or (fn.endswith(".cif") and len(fn) == 8):
                        locator = os.path.join(root, fn)
                        kwD = HashableDict({})
                        oL = [HashableDict({"locator": locator, "fmt": "mmcif", "kwargs": kwD})]
                        for mergeContentType in mergeContentTypes:
                            idCode = fn[:4] if fn and len(fn) >= 8 else None
                            mergeLocator = self.__getLocator(mergeContentType, idCode, checkExists=True) if idCode else None
                            if mergeLocator:
                                kwD = HashableDict({"marshalHelper": toCifWrapper})
                                oL.append(HashableDict({"locator": mergeLocator, "fmt": "xml", "kwargs": kwD}))
                        lObj = tuple(oL)
                        locatorObjList.append(lObj)
        return dataList, locatorObjList, []

    def __getEntryLocatorObjList(self, mergeContentTypes=None):
        return self.__fetchEntryLocatorObjList(self.__getRepoLocalPath("pdbx"), numProc=self.__numProc, mergeContentTypes=mergeContentTypes)

    def __fetchEntryLocatorObjList(self, topRepoPath, numProc=8, mergeContentTypes=None):
        """Get the path list for structure entries in the input repository"""
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s", ts)
        startTime = time.time()
        pathList = []
        try:
            dataList = []
            anL = "abcdefghijklmnopqrstuvwxyz0123456789"
            for a1 in anL:
                for a2 in anL:
                    hc = a1 + a2
                    dataList.append(hc)
                    hc = a2 + a1
                    dataList.append(hc)
            dataList = list(set(dataList))
            #
            optD = {}
            optD["topRepoPath"] = topRepoPath
            optD["mergeContentTypes"] = mergeContentTypes
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_entryLocatorObjWithMergeWorker")
            _, _, retLists, _ = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Locator object list length %d  in %.4f seconds", len(pathList), endTime0 - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(pathList)

    def _entryPathWorker(self, dataList, procName, optionsD, workingDir):
        """Return the list of entry file paths in the current repository."""
        _ = procName
        _ = workingDir
        topRepoPath = optionsD["topRepoPath"]
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, _, files in os.walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if (name.endswith(".cif.gz") and len(name) == 11) or (name.endswith(".cif") and len(name) == 8):
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def _compModelPathWorker(self, dataList, procName, optionsD, workingDir):
        """Return the list of computed-model entry file paths in the current computed-model storage area.

        Args:
            dataList (list): List of sub-directory paths to model files in the computed-model storage area, with paths starting at
                             the model source prefix (e.g., ["AF/XJ/E6/AF_AFA0A385XJE6F1.cif.gz", "MA/PC/05/MA_MABAKCEPC0534.cif.gz."])
            procName (str): worker process name
            optionsD (dict): dictionary of additional options that worker can access
            workingDir (str): path to working directory
        """
        _ = procName
        _ = workingDir
        topRepoPath = optionsD["topRepoPath"]
        modelFormat = optionsD["modelFormat"]
        pathList = []
        for modelId in dataList:
            modelSourcePrefix, firstDir, secondDir = modelId[0:2], modelId[-6:-4], modelId[-4:-2]
            internalModelName = modelId + modelFormat
            modelPathFromPrefixDir = os.path.join(modelSourcePrefix, firstDir, secondDir, internalModelName)
            pathList.append(os.path.join(topRepoPath, modelPathFromPrefixDir))
        return dataList, pathList, []

    def __getEntryPathList(self):
        return self.__fetchEntryPathList(self.__getRepoLocalPath("pdbx"), numProc=self.__numProc)

    def getObsoleteEntryPathList(self):
        return self.__fetchEntryPathList(self.__getRepoLocalPath("pdbx_obsolete"), numProc=self.__numProc)

    def __fetchEntryPathList(self, topRepoPath, numProc=8):
        """Get the path list for structure entries in the input repository"""
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s", ts)
        startTime = time.time()
        pathList = []
        try:
            dataList = []
            anL = "abcdefghijklmnopqrstuvwxyz0123456789"
            for a1 in anL:
                for a2 in anL:
                    hc = a1 + a2
                    dataList.append(hc)
                    hc = a2 + a1
                    dataList.append(hc)
            dataList = list(set(dataList))
            #
            optD = {}
            optD["topRepoPath"] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_entryPathWorker")
            _, _, retLists, _ = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds", len(pathList), endTime0 - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(pathList)

    def __getBirdPathList(self):
        return self.__fetchBirdPathList(self.__getRepoLocalPath("bird"))

    def __fetchBirdPathList(self, topRepoPath):
        """Return the list of definition file paths in the current repository.

        List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, _, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRD_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return self.__applyLimit(pathList)

    def __getBirdFamilyPathList(self):
        return self.__fetchBirdFamilyPathList(self.__getRepoLocalPath("bird_family"))

    def __fetchBirdFamilyPathList(self, topRepoPath):
        """Return the list of definition file paths in the current repository.

        List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, _, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("FAM_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return self.__applyLimit(pathList)

    def __getBirdChemCompPathList(self):
        return self.__fetchBirdChemCompPathList(self.__getRepoLocalPath("bird_chem_comp"))

    def __fetchBirdChemCompPathList(self, topRepoPath):
        """Return the list of definition file paths in the current repository.

        List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, _, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRDCC_") and name.endswith(".cif") and len(name) <= 16:
                        pth = os.path.join(root, name)
                        sd[int(name[6:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        return self.__applyLimit(pathList)

    def __applyLimit(self, itemList):
        logger.debug("Length of item list %d (limit %r)", len(itemList), self.__fileLimit)
        if self.__fileLimit:
            return itemList[: self.__fileLimit]
        else:
            return itemList

    def __buildFamilyIndex(self):
        """Using information from the PRD family definition:
        #
        loop_
        _pdbx_reference_molecule_list.family_prd_id
        _pdbx_reference_molecule_list.prd_id
            FAM_000010 PRD_000041
            FAM_000010 PRD_000042
            FAM_000010 PRD_000043
            FAM_000010 PRD_000044
            FAM_000010 PRD_000048
            FAM_000010 PRD_000049
            FAM_000010 PRD_000051
        #
        """
        prdD = {}
        try:
            pthL = self.getLocatorPaths(self.__getLocatorList("bird_family"))
            for pth in pthL:
                containerL = self.__mU.doImport(pth, fmt="mmcif")
                for container in containerL:
                    catName = "pdbx_reference_molecule_list"
                    if container.exists(catName):
                        catObj = container.getObj(catName)
                        for ii in range(catObj.getRowCount()):
                            familyPrdId = catObj.getValue(attributeName="family_prd_id", rowIndex=ii)
                            prdId = catObj.getValue(attributeName="prd_id", rowIndex=ii)
                            if prdId in prdD:
                                logger.debug("duplicate prdId in family index %s %s", prdId, familyPrdId)
                            prdD[prdId] = {"familyPrdId": familyPrdId, "c": container}
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        return prdD

    def __buildBirdCcIndex(self, idCodeList=None):
        """Using information from the PRD pdbx_reference_molecule category to
        index the BIRDs corresponding small molecule correspondences

        """
        prdD = {}
        ccPathD = {}
        prdStatusD = {}
        try:
            ccPathL = self.getLocatorPaths(self.__getLocatorList("chem_comp", inputIdCodeList=idCodeList))
            logger.debug("ccPathL: %r", ccPathL)
            ccPathD = {}
            for ccPath in ccPathL:
                _, fn = os.path.split(ccPath)
                ccId, _ = os.path.splitext(fn)
                ccPathD[ccId] = ccPath
            logger.info("Chemical component path list (%d)", len(ccPathD))
            # logger.info("Chemical component path list: %r", ccPathD)
            #
            pthL = self.getLocatorPaths(self.__getLocatorList("bird", inputIdCodeList=idCodeList))
            logger.info("BIRD path list (%d)", len(pthL))
            # logger.info("BIRD path list: %r", pthL)
            #
            for pth in pthL:
                containerL = self.__mU.doImport(pth, fmt="mmcif")
                for container in containerL:
                    catName = "pdbx_reference_molecule"
                    if container.exists(catName):
                        catObj = container.getObj(catName)
                        ii = 0
                        prdId = catObj.getValue(attributeName="prd_id", rowIndex=ii)
                        relStatus = catObj.getValue(attributeName="release_status", rowIndex=ii)
                        prdStatusD[prdId] = relStatus
                        if relStatus != "REL":
                            continue
                        prdRepType = catObj.getValue(attributeName="represent_as", rowIndex=ii)
                        logger.debug("represent as %r", prdRepType)
                        if prdRepType in ["single molecule"]:
                            ccId = catObj.getValueOrDefault(attributeName="chem_comp_id", rowIndex=ii, defaultValue=None)
                            # prdId = catObj.getValue(attributeName="prd_id", rowIndex=ii)
                            logger.debug("mapping prdId %r ccId %r", prdId, ccId)
                            if ccId and ccId in ccPathD:
                                prdD[prdId] = {"ccId": ccId, "ccPath": ccPathD[ccId]}
                                ccPathD[ccPathD[ccId]] = {"ccId": ccId, "prdId": prdId}
                            else:
                                logger.warning("Missing ccId %r referenced in BIRD %r", ccId, prdId)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        logger.info("Candidate Chemical Components (%d) BIRDS (%d) BIRD status details (%d)", len(prdD), len(ccPathD), len(prdStatusD))
        return prdD, ccPathD, prdStatusD

    # -
    def mergeBirdAndChemCompRefData(self):
        # JDW note that this merging procedure expects access to all reference data -
        # Use this method when:  self.__fileLimit = None
        prdSmallMolCcD, ccPathD, prdStatusD = self.__buildBirdCcIndex()
        logger.info("PRD to CCD index length %d CCD map path length %d", len(prdSmallMolCcD), len(ccPathD))
        outputPathList = self.__mergeBirdRefData(prdSmallMolCcD, prdStatusD)
        #
        ccOutputPathList = []
        if self.__discoveryMode == 'remote':
            for pth in self.__getChemCompUriList(idCodeList=None):
                ccp = pth[0]['locator']
                if ccp not in ccPathD:
                    ccOutputPathList.append(ccp)
        else:
            ccOutputPathList = [pth for pth in self.__getChemCompPathList() if pth not in ccPathD]
        #
        outputPathList.extend(ccOutputPathList)
        logger.info("Total cc paths: %d", len(ccOutputPathList))
        logger.info("Total bird_chem_comp paths: %d", len(outputPathList))
        #
        return outputPathList

    def mergeBirdAndChemCompRefDataWithInput(self, idCodeList=None):
        # Use this method when:  self.__fileLimit != None
        idCodeList = idCodeList if idCodeList else []
        prdSmallMolCcD, ccPathD, prdStatusD = self.__buildBirdCcIndex(idCodeList=idCodeList)
        logger.info("PRD to CCD index length %d CCD map path length %d", len(prdSmallMolCcD), len(ccPathD))
        outputPathList = self.__mergeBirdRefData(prdSmallMolCcD, prdStatusD, idCodeList=idCodeList)
        #
        ccOutputPathList = []
        if self.__discoveryMode == 'remote':
            for pth in self.__getChemCompUriList(idCodeList=idCodeList):
                ccp = pth[0]['locator']
                if ccp not in ccPathD:
                    ccOutputPathList.append(ccp)
        else:
            ccOutputPathList = [pth for pth in self.__getChemCompPathList() if pth not in ccPathD]
        #
        outputPathList.extend(ccOutputPathList)
        logger.info("Total cc paths: %d", len(ccOutputPathList))
        logger.info("Total bird_chem_comp paths: %d", len(outputPathList))
        #
        return outputPathList

    def __mergeBirdRefData(self, prdSmallMolCcD, prdStatusD, idCodeList=None):
        """Consolidate all of the bird reference data in a single container.

        If the BIRD is a 'small molecule' type then also merge with the associated CC definition.

        Store the merged data in the REPO_UTIL cache path and ...

        Return a path list for the consolidated data files -

        """
        outPathList = []
        iSkipUnreleased = 0
        try:
            birdPathList = self.getLocatorPaths(self.__getLocatorList("bird", inputIdCodeList=idCodeList))
            birdPathD = {}
            for birdPath in birdPathList:
                _, fn = os.path.split(birdPath)
                prdId, _ = os.path.splitext(fn)
                birdPathD[prdId] = birdPath
            #
            logger.info("BIRD path length %d", len(birdPathD))
            logger.debug("BIRD keys %r", list(birdPathD.keys()))
            birdCcPathList = self.getLocatorPaths(self.__getLocatorList("bird_chem_comp", inputIdCodeList=idCodeList))
            birdCcPathD = {}
            for birdCcPath in birdCcPathList:
                _, fn = os.path.split(birdCcPath)
                prdCcId, _ = os.path.splitext(fn)
                prdId = "PRD_" + prdCcId[6:]
                birdCcPathD[prdId] = birdCcPath
            #
            logger.info("BIRDCC path length %d", len(birdCcPathD))
            logger.debug("BIRD CC keys %r", list(birdCcPathD.keys()))
            fD = self.__buildFamilyIndex()
            logger.info("BIRD Family index length %d", len(fD))
            logger.debug("Family index keys %r", list(fD.keys()))
            logger.info("PRD to CCD small mol index length %d", len(prdSmallMolCcD))
            #
            for prdId in birdPathD:
                if prdId in prdStatusD and prdStatusD[prdId] != "REL":
                    logger.debug("Skipping BIRD with non-REL status %s", prdId)
                    iSkipUnreleased += 1
                    continue
                fp = os.path.join(self.__cachePath, prdId + ".cif")
                logger.debug("Export cache path is %r", fp)
                #
                pth2 = birdPathD[prdId]
                cL = self.__mU.doImport(pth2, fmt="mmcif")
                cFull = cL[0]
                logger.debug("Got Bird %r", cFull.getName())
                #
                # --- JDW
                # add missing one_letter_codes item
                if cFull.exists("pdbx_reference_entity_sequence") and cFull.exists("pdbx_reference_entity_poly_seq"):
                    aaDict3 = {
                        "ALA": "A",
                        "ARG": "R",
                        "ASN": "N",
                        "ASP": "D",
                        "ASX": "B",
                        "CYS": "C",
                        "GLN": "Q",
                        "GLU": "E",
                        "GLX": "Z",
                        "GLY": "G",
                        "HIS": "H",
                        "ILE": "I",
                        "LEU": "L",
                        "LYS": "K",
                        "MET": "M",
                        "PHE": "F",
                        "PRO": "P",
                        "SER": "S",
                        "THR": "T",
                        "TRP": "W",
                        "TYR": "Y",
                        "VAL": "V",
                        "PYL": "O",
                        "SEC": "U",
                    }
                    catObj = cFull.getObj("pdbx_reference_entity_sequence")
                    if not catObj.hasAttribute("one_letter_codes"):
                        logger.debug("adding one letter codes for %r", prdId)
                        seqObj = cFull.getObj("pdbx_reference_entity_poly_seq")
                        seqD = {}
                        for jj in range(0, seqObj.getRowCount()):
                            entityId = seqObj.getValue("ref_entity_id", jj)
                            monId = seqObj.getValue("mon_id", jj)
                            seqD.setdefault(entityId, []).append(monId)
                        #
                        logger.debug("seqD %r", seqD)
                        catObj.appendAttribute("one_letter_codes")
                        for ii in range(catObj.getRowCount()):
                            entityId = catObj.getValue("ref_entity_id", ii)
                            if entityId in seqD:
                                ttL = [aaDict3[tt] if tt in aaDict3 else "X" for tt in seqD[entityId]]
                                catObj.setValue("".join(ttL), "one_letter_codes", ii)
                            else:
                                logger.error("%r missing sequence for entity %r", prdId, entityId)
                # ---
                #
                ccBird = None
                ccD = None
                if prdId in prdSmallMolCcD:
                    try:
                        pthCc = prdSmallMolCcD[prdId]["ccPath"]
                        cL = self.__mU.doImport(pthCc, fmt="mmcif")
                        ccD = cL[0]
                    except Exception as e:
                        logger.error("(%s) failed getting path %r: %r", prdId, pthCc, str(e))
                    #
                elif prdId in birdCcPathD:
                    try:
                        pth1 = birdCcPathD[prdId]
                        c1L = self.__mU.doImport(pth1, fmt="mmcif")
                        ccBird = c1L[0]
                    except Exception as e:
                        logger.error("(%s) Failed getting path %r: %r", prdId, pth1, str(e))
                    #
                cFam = None
                if prdId in fD:
                    cFam = fD[prdId]["c"]
                    logger.debug("Got cFam %r", cFam.getName())
                #
                if ccD:
                    for catName in ccD.getObjNameList():
                        cFull.append(ccD.getObj(catName))
                #
                if ccBird:
                    for catName in ccBird.getObjNameList():
                        cFull.append(ccBird.getObj(catName))
                if cFam:
                    for catName in cFam.getObjNameList():
                        cFull.append(cFam.getObj(catName))
                #
                self.__mU.doExport(fp, [cFull], fmt="mmcif")
                outPathList.append(fp)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        #
        logger.info("Merged BIRD/Family/CC path length %d (skipped non-released %d)", len(outPathList), iSkipUnreleased)
        return outPathList
        #

    def __getCompModelPathList(self, idCodeList=None, fmt="mmcif"):
        return self.__fetchModelPathList(self.__getRepoLocalPath("pdbx_comp_model_core"), idCodeList=idCodeList, fmt=fmt, numProc=self.__numProc)

    def __fetchModelPathList(self, topRepoPath, idCodeList=None, fmt="mmcif", numProc=8):
        """Get the path list for computational models in the input cached model repository

        TO-DO: Add check of cache file to see if it changed between the last time data was uploaded, and if so, then upload new models

        File name template is:  <topRepoPath>/<2-char source>/<hash>/<hash>/*.cif.gz
        """
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.info("Starting at %s", ts)
        logger.info("Computed-models topRepoPath: %s", topRepoPath)
        startTime = time.time()
        #
        idCodeList = idCodeList if idCodeList else []
        pathList = []
        try:
            optD = {}
            optD["topRepoPath"] = topRepoPath
            optD["modelFormat"] = ".bcif.gz" if "bcif" in fmt.lower() else ".cif.gz"
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_compModelPathWorker")
            _, _, retLists, _ = mpu.runMulti(dataList=idCodeList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds", len(pathList), endTime0 - startTime)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return self.__applyLimit(pathList)

    def __getIhmDevPathList(self):
        return self.__fetchIhmDevPathList(self.__getRepoLocalPath("ihm_dev"))

    def __fetchIhmDevPathList(self, topRepoPath):
        """Return the list of I/HM entries in the current repository.

        File name template is: PDBDEV_0000 0020_model_v1-0.cif.gz

        List is ordered in increasing PRDDEV numerical code.
        """
        pathList = []
        logger.debug("Searching path %r", topRepoPath)
        try:
            sd = {}
            for root, _, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PDBDEV_") and name.endswith(".cif.gz") and len(name) <= 50:
                        pth = os.path.join(root, name)
                        sd[int(name[7:15])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing search in %r with %s", topRepoPath, str(e))
        #
        return self.__applyLimit(pathList)



##
#  File:  CurrentHoldingsProvider.py
#  Date:  18-May-2021 jdw
#
#  Updates:
#   12-Jun-2023  dwp Set useCache default to False to force redownloading of holdings files
#    1-Jul-2024  dwp Stop populating "2fo-fc Map" and "fo-fc Map" content types (DSN6 maps), and
#                    only include "Map Coefficients" (MTZ map coefficients) in repository holdings
#    9-Sep-2024  dwp Always defer to loading holdings data from remote (rather than storing it locally);
#                    Add validation coefficients to list of repository_content_types
#
##
"""Provide inventory of current repository content.
"""

import datetime
import logging
import os.path

import pytz

from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.struct.EntryInfoProvider import EntryInfoProvider

logger = logging.getLogger(__name__)


class CurrentHoldingsProvider(object):
    """Provide inventory of current repository content."""

    def __init__(self, cachePath, useCache=False, **kwargs):
        self.__cachePath = cachePath
        self.__dirPath = os.path.join(cachePath, "holdings")
        self.__storeCache = kwargs.get("storeCache", False)
        #
        edMapsLocator = kwargs.get("edmapsLocator", "https://raw.githubusercontent.com/rcsb/py-rcsb_exdb_assets/master/fall_back/edmaps.json")
        #
        baseUrl = kwargs.get("holdingsTargetUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        fallbackUrl = kwargs.get("holdingsFallbackUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        #
        entryUrlContent = os.path.join(baseUrl, "current_file_holdings.json.gz")
        entryUrlFallbackContent = os.path.join(fallbackUrl, "current_file_holdings.json.gz")
        entryUrlIds = os.path.join(baseUrl, "released_structures_last_modified_dates.json.gz")
        entryUrlFallbackIds = os.path.join(fallbackUrl, "released_structures_last_modified_dates.json.gz")
        #
        refdataUrlIds = os.path.join(baseUrl, "refdata_id_list.json.gz")
        refdataUrlFallbackIds = os.path.join(fallbackUrl, "refdata_id_list.json.gz")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)

        self.__invD = self.__reloadEntryContent(entryUrlContent, entryUrlFallbackContent, edMapsLocator, self.__dirPath, useCache=useCache)
        self.__idD = self.__reloadEntryIds(entryUrlIds, entryUrlFallbackIds, self.__dirPath, useCache=useCache)
        self.__refD = self.__reloadRefdataIds(refdataUrlIds, refdataUrlFallbackIds, self.__dirPath, useCache=useCache)
        # EntryInfoProvider must be cached before this class is invoked -
        self.__eiP = EntryInfoProvider(cachePath=self.__cachePath, useCache=True)
        ok = self.__eiP.testCache()
        if not ok:
            self.__eiP = None

    def testCache(self, minCount=220000):
        logger.info("Inventory length cD (%d) id list (%d)", len(self.__invD), len(self.__idD))
        # JDW - restore consistency checks
        # if len(self.__invD) > minCount and len(self.__idD) > minCount and len(self.__invD) == len(self.__idD):
        if len(self.__invD) > minCount and len(self.__idD) > minCount:
            return True
        return False

    def hasEntryContentType(self, entryId, contentType):
        """Return if the current content types is available for the input entry identifier"""
        try:
            return contentType in self.__invD[entryId.upper()]
        except Exception as e:
            logger.exception("Failing for %r with %s", entryId, str(e))
        return False

    def getEntryContentTypes(self, entryId):
        """Return the current content types for the input entry identifier"""
        try:
            return sorted(self.__invD[entryId.upper()].keys())
        except Exception as e:
            logger.exception("Failing for %r with %s", entryId, str(e))
        return []

    def getAllContentTypes(self):
        """Return the all current content types for the repository"""
        entryId = None
        try:
            tS = set()
            for entryId, tD in self.__invD.items():
                for ky in tD:
                    tS.add(ky)
            return sorted(tS)
        except Exception as e:
            logger.exception("Failing for %r with %s", entryId, str(e))
        return []

    def getEntryContentTypePathList(self, entryId, contentType):
        """Return the current content types for the input entry identifier"""
        try:
            return self.__invD[entryId.upper()][contentType]
        except Exception as e:
            logger.debug("Failing for %r %r with %s", entryId, contentType, str(e))
        return []

    def getEntryInventory(self):
        """Return the current inventory dictionary"""
        try:
            return self.__invD
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return {}

    def getEntryIdList(self, afterDateTimeStamp=None):
        """Return the ID code list or optionally IDs changed after the input time stamp.

        Args:
            afterDateTimeStamp (str, optional): ISO format date time stamp. Defaults to None.
        """
        try:
            if afterDateTimeStamp:
                dt = datetime.datetime.fromisoformat(afterDateTimeStamp).replace(tzinfo=pytz.utc)
                return [k for k, v in self.__idD.items() if v > dt]
            else:
                return list(self.__idD.keys())
        except Exception as e:
            logger.error("Failing with %s", str(e))
        #
        return []

    def getRcsbContentAndAssemblies(self):
        return self.__assembleEntryContentTypes(self.__invD)

    # --- reference data ---
    def getBirdIdList(self):
        return self.__getRefIdListByType("BIRD")

    def getBirdChemCompIdList(self):
        return self.__getRefIdListByType("BIRD CC")

    def getBirdFamilyIdList(self):
        return self.__getRefIdListByType("BIRD Family")

    def getChemCompIdList(self):
        return self.__getRefIdListByType("CC")

    def __getRefIdListByType(self, refType):
        idL = []
        try:
            for rId, tup in self.__refD.items():
                if tup and tup[0] == refType:
                    idL.append(rId)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return idL

    # ---
    def getStatusDetails(self):
        retD = {}
        for entryId in self.__invD:
            retD[entryId] = {"status": "CURRENT", "status_code": "REL"}
        return retD

    def hasValidationReportData(self, entryId):
        return self.__hasValidationReportData(self.__invD, entryId)

    def __hasValidationReportData(self, invD, entryId):
        if entryId.upper() in invD:
            tD = invD[entryId.upper()]
            if "validation_report" in tD:
                for pth in tD["validation_report"]:
                    if pth[-7:] == ".xml.gz":
                        return True
        return False

    def __assembleEntryContentTypes(self, invD):
        # Mapping between repository content types and those used by RCSB.org
        contentTypeD = {
            "mtz_map_coefficients": "Map Coefficients",
            # "2fofc Map": "2fo-fc Map",
            # "fofc Map": "fo-fc Map",
            "assembly_mmcif": "assembly mmCIF",
            "assembly_pdb": "assembly PDB",
            "combined_nmr_data_nef": "Combined NMR data (NEF)",
            "combined_nmr_data_nmr-star": "Combined NMR data (NMR-STAR)",
            "mmcif": "entry mmCIF",
            "nmr_chemical_shifts": "NMR chemical shifts",
            "nmr_restraints_v1": "NMR restraints V1",
            "nmr_restraints_v2": "NMR restraints V2",
            "pdb": "entry PDB",
            "pdb_bundle": "entry PDB bundle",
            "pdbml": "entry PDBML",
            # "pdbml_extatom":  "entry PDBML",
            # "pdbml_noatom": "entry PDBML",
            "structure_factors": "structure factors",
            # Missing mappings
            # "validation report",
            # "validation slider image"
            # "validation 2fo-fc coefficients"
            # "validation fo-fc coefficients"
            # "Map Coefficients",
            # "FASTA sequence",
        }
        #
        noPolymerL = self.__eiP.getEntriesByPolymerEntityCount(count=0) if self.__eiP else []
        logger.info("Entries missing polymers (%d)", len(noPolymerL))
        # for id in noPolymerL:
        #    logger.info("id: %s", id)
        ctD = {}
        assemD = {}
        for entryId, tD in invD.items():
            assemS = set()
            if entryId not in noPolymerL:
                ctD.setdefault(entryId, []).append("FASTA sequence")
            for contentType, pthL in tD.items():
                if contentType in contentTypeD:
                    ctD.setdefault(entryId, []).append(contentTypeD[contentType])
                if contentType == "validation_report":
                    # "/pdb/validation_reports/01/201l/201l_full_validation.pdf.gz"
                    # "/pdb/validation_reports/01/201l/201l_multipercentile_validation.png.gz"
                    # "/pdb/validation_reports/01/201l/201l_multipercentile_validation.svg.gz"
                    # "/pdb/validation_reports/01/201l/201l_validation.pdf.gz"
                    # "/pdb/validation_reports/01/201l/201l_validation.xml.gz"
                    # "/pdb/validation_reports/01/201l/201l_validation_2fo-fc_map_coef.cif.gz"
                    # "/pdb/validation_reports/01/201l/201l_validation_fo-fc_map_coef.cif.gz"
                    for pth in pthL:
                        # Use "_full_validation.pdf.gz" instead of just ".pdf.gz" to avoid re-appending for non-full "_validation.pdf.gz" file
                        if "full_validation.pdf.gz" in pth:
                            ctD.setdefault(entryId, []).append("validation report")
                        elif "validation.svg.gz" in pth:
                            ctD.setdefault(entryId, []).append("validation slider image")
                        elif "validation.cif.gz" in pth:
                            ctD.setdefault(entryId, []).append("validation data mmCIF")
                        elif "validation_2fo-fc_map_coef.cif.gz" in pth:
                            ctD.setdefault(entryId, []).append("validation 2fo-fc coefficients")
                        elif "validation_fo-fc_map_coef.cif.gz" in pth:
                            ctD.setdefault(entryId, []).append("validation fo-fc coefficients")
                if contentType == "assembly_mmcif":
                    # "/pdb/data/biounit/mmCIF/divided/a0/7a09-assembly1.cif.gz"
                    for pth in pthL:
                        fn = os.path.basename(pth)
                        aId = fn.split(".")[0].split("-")[1].replace("assembly", "")
                        assemS.add(aId)

                if contentType == "assembly_pdb":
                    # "/pdb/data/biounit/coordinates/divided/02/302d.pdb1.gz"
                    for pth in pthL:
                        fn = os.path.basename(pth)
                        aId = fn.split(".")[1].replace("pdb", "")
                        assemS.add(aId)
            assemD[entryId] = list(assemS)
        return ctD, assemD

    def __addMapContents(self, invD, mapD):
        """Add map content types to the entry inventory

        Args:
            invD (dict): entry inventory
            mapD (dict): map inventory

            Example - edmaps.json {"6aok": {"2fofc": "true", "fofc": "true"}, "4ih7": {"2fofc": "true", "fofc": "true"}, ...}
        """
        for entryId in invD:
            if entryId.lower() in mapD:
                tD = mapD[entryId.lower()]
                if "2fofc" in tD and tD["2fofc"] == "true":
                    invD[entryId]["mtz_map_coefficients"] = []
        return invD

    def __reloadEdmapContent(self, edmapsLocator, dirPath):
        invD = {}
        try:
            invD = self.__mU.doImport(edmapsLocator, fmt="json")
            logger.info("Loaded edmaps inventory from %s (%r)", edmapsLocator, len(invD))
            #
            if self.__storeCache:
                fU = FileUtil()
                fn = fU.getFileName(edmapsLocator)
                fp = os.path.join(dirPath, fn)
                self.__mU.mkdir(dirPath)
                ok = fU.get(edmapsLocator, fp)
                logger.info("Fetch edmaps inventory from %s to %s (%r)", edmapsLocator, fp, ok)
        except Exception as e:
            logger.exception("Failing for %r with %s", edmapsLocator, str(e))
        return invD

    def __reloadEntryContent(self, urlTarget, urlFallbackTarget, edMapsLocator, dirPath, useCache=True):
        invD = {}
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        fp = os.path.join(dirPath, fn)
        self.__mU.mkdir(dirPath)
        #
        if self.__storeCache and useCache and self.__mU.exists(fp):
            invD = self.__mU.doImport(fp, fmt="json")
            logger.info("Reading cached inventory (%d)", len(invD))
        else:
            invD = self.__mU.doImport(urlTarget, fmt="json")
            logger.info("Loaded inventory from %s (%r)", urlTarget, len(invD))
            if len(invD) == 0:
                invD = self.__mU.doImport(urlFallbackTarget, fmt="json")
                logger.info("Loaded fallback inventory from %s (%r)", urlFallbackTarget, len(invD))
            mapD = self.__reloadEdmapContent(edMapsLocator, self.__dirPath)
            invD = self.__addMapContents(invD, mapD)
            del mapD
            #
            # previous method - save file locally
            if self.__storeCache:
                logger.info("Fetch inventory from %s", urlTarget)
                ok = fU.get(urlTarget, fp)
                if not ok:
                    ok = fU.get(urlFallbackTarget, fp)
                if ok:
                    ofp = fp[:-3] if fp.endswith(".gz") else fp
                    ok = self.__mU.doExport(ofp, invD, fmt="json", indent=3)
                    if fp.endswith(".gz"):
                        logger.info("Updating the current entry contents (%r) in %r", ok, fp)
                        fU.compress(ofp, fp)
        return invD

    def __reloadEntryIds(self, urlTarget, urlFallbackTarget, dirPath, useCache=True):
        tD = {}
        idD = {}
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        fp = os.path.join(dirPath, fn)
        self.__mU.mkdir(dirPath)
        #
        if self.__storeCache and useCache and self.__mU.exists(fp):
            tD = self.__mU.doImport(fp, fmt="json")
            logger.info("Reading cached IDs list (%d)", len(tD))
        else:
            tD = self.__mU.doImport(urlTarget, fmt="json")
            logger.info("Loaded ID list from %s (%r)", urlTarget, len(tD))
            if len(tD) == 0:
                tD = self.__mU.doImport(urlFallbackTarget, fmt="json")
                logger.info("Loaded fallback ID list from %s (%r)", urlFallbackTarget, len(tD))
        #
        if self.__storeCache:
            logger.info("Fetch ID list from %s", urlTarget)
            ok = fU.get(urlTarget, fp)
            if not ok:
                ok = fU.get(urlFallbackTarget, fp)
        #
        for k, v in tD.items():
            try:
                idD[k] = datetime.datetime.fromisoformat(v)
            except Exception as e:
                logger.error("Date processing failing for %r %r with %s", k, v, str(e))
        #
        sTupL = sorted(idD.items(), key=lambda item: item[1])
        return {k: v for k, v in sTupL}

    def __reloadRefdataIds(self, urlTarget, urlFallbackTarget, dirPath, useCache=True):
        tD = {}
        idD = {}
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        fp = os.path.join(dirPath, fn)
        self.__mU.mkdir(dirPath)
        #
        if self.__storeCache and useCache and self.__mU.exists(fp):
            tD = self.__mU.doImport(fp, fmt="json")
            logger.info("Reading cached IDs list (%d)", len(tD))
        else:
            tD = self.__mU.doImport(urlTarget, fmt="json")
            logger.info("Loaded ID list from %s (%r)", urlTarget, len(tD))
            if len(tD) == 0:
                tD = self.__mU.doImport(urlFallbackTarget, fmt="json")
                logger.info("Loaded fallback ID list from %s (%r)", urlFallbackTarget, len(tD))
        #
        if self.__storeCache:
            logger.info("Fetch ID list from %s", urlTarget)
            ok = fU.get(urlTarget, fp)
            if not ok:
                ok = fU.get(urlFallbackTarget, fp)
        #
        for k, v in tD.items():
            try:
                idD[k] = (v["content_type"], datetime.datetime.fromisoformat(v["last_modified_date"]))
            except Exception as e:
                logger.error("Date processing failing for %r %r with %s", k, v, str(e))
        return idD



##
#  File:  UnreleasedHoldingsProvider.py
#  Date:  24-Sep-2021 jdw
#
#  Updates:
#   12-Jun-2023  dwp Set useCache default to False to force redownloading of holdings files
#    9-Sep-2024  dwp Always defer to loading holdings data from remote (rather than storing it locally)
#
##
"""Provide an inventory of unreleased repository content.
"""

import logging
import os.path

import dateutil.parser
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class UnreleasedHoldingsProvider(object):
    """Provide an inventory of unreleased repository content."""

    def __init__(self, cachePath, useCache=False, **kwargs):
        self.__dirPath = os.path.join(cachePath, "holdings")
        self.__storeCache = kwargs.get("storeCache", False)
        self.__filterType = kwargs.get("filterType", "")
        self.__assignDates = "assign-dates" in self.__filterType
        baseUrl = kwargs.get("holdingsTargetUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        fallbackUrl = kwargs.get("holdingsFallbackUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        #
        urlTarget = os.path.join(baseUrl, "unreleased_entries.json.gz")
        urlFallbackTarget = os.path.join(fallbackUrl, "unreleased_entries.json.gz")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__invD = self.__reload(urlTarget, urlFallbackTarget, self.__dirPath, useCache=useCache)

    def testCache(self, minCount=5000):
        logger.info("Inventory length cD (%d)", len(self.__invD))
        if len(self.__invD) > minCount:
            return True
        return False

    def getStatusCode(self, entryId):
        """Return the status code for the unreleased entry"""
        try:
            return self.__invD[entryId.upper()]["status_code"]
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return None

    def getUnreleasedInfo(self, entryId):
        """Return the dictionary describing the details for this unreleased entry"""
        try:
            return self.__invD[entryId.upper()]
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return {}

    def getInventory(self):
        """Return the unreleased inventory dictionary"""
        try:
            return self.__invD
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return {}

    def __reload(self, urlTarget, urlFallbackTarget, dirPath, useCache=True):
        invD = {}
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        fp = os.path.join(dirPath, fn)
        self.__mU.mkdir(dirPath)
        #
        if self.__storeCache and useCache and self.__mU.exists(fp):
            invD = self.__mU.doImport(fp, fmt="json")
            logger.debug("Reading cached inventory (%d)", len(invD))
        else:
            invD = self.__mU.doImport(urlTarget, fmt="json")
            logger.info("Loaded inventory from %s (%r)", urlTarget, len(invD))
            if len(invD) == 0:
                invD = self.__mU.doImport(urlFallbackTarget, fmt="json")
                logger.info("Loaded fallback inventory from %s (%r)", urlFallbackTarget, len(invD))
        #
        if self.__storeCache:
            logger.info("Fetch inventory from %s", urlTarget)
            ok = fU.get(urlTarget, fp)
            if not ok:
                ok = fU.get(urlFallbackTarget, fp)
        #
        return invD

    def getStatusDetails(self, curD):
        sD = {}
        for entryId, tD in self.__invD.items():
            if entryId not in curD and tD["status_code"] in ["AUCO", "AUTH", "HOLD", "HPUB", "POLC", "PROC", "REFI", "REPL", "WAIT", "WDRN"]:
                sD[entryId] = {"status": "UNRELEASED", "status_code": tD["status_code"]}
        return sD

    def getRcsbUnreleasedData(self):
        return self.__mapAttributes(self.__invD)

    def __mapAttributes(self, invD):
        #
        mapD = {
            "deposit_date": "deposit_date",
            "title": "title",
            "deposit_date_nmr_constraints": "deposit_date_nmr_restraints",
            "prerelease_sequence_available_flag": "prerelease_sequence_available_flag",
            "deposit_date_structure_factors": "deposit_date_structure_factors",
            "hold_date_structure_factors": "hold_date_structure_factors",
            "author_prerelease_sequence_status": "author_prerelease_sequence_status",
            "deposit_date_coordinates": "deposit_date_coordinates",
            "hold_date_coordinates": "hold_date_coordinates",
            "hold_date_nmr_constraints": "hold_date_nmr_restraints",
            "deposition_authors": "audit_authors",
            "status_code": "status_code",
        }
        dateFields = [
            "deposit_date",
            "deposit_date_coordinates",
            "deposit_date_structure_factors",
            "hold_date_structure_factors",
            "deposit_date_nmr_restraints",
            "hold_date_nmr_restraints",
            "release_date",
            "hold_date_coordinates",
        ]
        retD = {}
        for entryId, tD in invD.items():
            qD = {}
            for iky, oky in mapD.items():
                if iky not in tD:
                    continue
                if self.__assignDates and oky in dateFields:
                    qD[oky] = dateutil.parser.parse(tD[iky])
                elif oky in dateFields:
                    qD[oky] = tD[iky][:10]
                else:
                    qD[oky] = tD[iky]
                #
                if iky == "author_prerelease_sequence_status":
                    qD[oky] = str(qD[oky]).strip().replace("REALEASE", "RELEASE")
            retD[entryId] = qD
        seqD = {}
        for entryId, tD in invD.items():
            if "prerelease_sequence" in tD:
                seqD[entryId] = tD["prerelease_sequence"]
        #
        return retD, seqD



##
#  File:  RemovedHoldingsProvider.py
#  Date:  18-May-2021 jdw
#
#  Updates:
#   18-Apr-2022  dwp Update getSupersededBy method to recursively return all superseded entries
#   12-Jun-2023  dwp Set useCache default to False to force redownloading of holdings files
#    9-Sep-2024  dwp Always defer to loading holdings data from remote (rather than storing it locally)
##
"""Provide an inventory of removed repository content.
"""
import logging
import os.path

import dateutil.parser
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class RemovedHoldingsProvider(object):
    """Provide an inventory of removed repository content."""

    def __init__(self, cachePath, useCache=False, **kwargs):
        self.__cachePath = cachePath
        self.__dirPath = os.path.join(self.__cachePath, "holdings")
        self.__storeCache = kwargs.get("storeCache", False)
        self.__filterType = kwargs.get("filterType", "")
        self.__assignDates = "assign-dates" in self.__filterType
        baseUrl = kwargs.get("holdingsTargetUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        fallbackUrl = kwargs.get("holdingsFallbackUrl", "https://files.wwpdb.org/pub/pdb/holdings")
        #
        urlTarget = os.path.join(baseUrl, "all_removed_entries.json.gz")
        urlFallbackTarget = os.path.join(fallbackUrl, "all_removed_entries.json.gz")
        #
        self.__mU = MarshalUtil(workPath=self.__dirPath)
        self.__invD = self.__reload(urlTarget, urlFallbackTarget, self.__dirPath, useCache=useCache)

    def testCache(self, minCount=1000):
        logger.info("Inventory length cD (%d)", len(self.__invD))
        if len(self.__invD) > minCount:
            return True
        return False

    def getStatusCode(self, entryId):
        """Return the status code for the removed entry"""
        try:
            return self.__invD[entryId.upper()]["status_code"]
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return None

    def getSupersededBy(self, entryId):
        """Return the superseding entry ids"""
        try:
            sL = []
            if isinstance(self.__invD[entryId.upper()]["superseded_by"], str):
                sL = [self.__invD[entryId.upper()]["superseded_by"]]
            else:
                sL = self.__invD[entryId.upper()]["superseded_by"]
            if len(sL) > 0:
                for recursiveEntry in sL:
                    if isinstance(self.__invD[recursiveEntry.upper()]["superseded_by"], str):
                        sL = sL + [self.__invD[recursiveEntry.upper()]["superseded_by"]]
                    elif isinstance(self.__invD[recursiveEntry.upper()]["superseded_by"], list):
                        sL = sL + self.__invD[recursiveEntry.upper()]["superseded_by"]
                    else:
                        break
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return sL

    def getRemovedEntries(self):
        return list(self.__invD.keys())

    def getEntryByStatus(self, statusCode):
        """Return the entry codes for removed entries with the input status code"""
        try:
            return [entryId for entryId, vD in self.__invD.items() if "status_code" in vD and vD["status_code"] == statusCode]
        except Exception as e:
            logger.debug("Failing for %r with %s", statusCode, str(e))
        return []

    def getRemovedInfo(self, entryId):
        """Return the dictionary describing the details for this removed entry"""
        try:
            return self.__invD[entryId.upper()]
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return {}

    def getContentTypes(self, entryId):
        """Return the removed content types for the input entry identifier"""
        try:
            return sorted(self.__invD[entryId.upper()]["content_type"].keys())
        except Exception as e:
            logger.debug("Failing for %r with %s", entryId, str(e))
        return []

    def getContentTypePathList(self, entryId, contentType):
        """Return the removed content types for the input entry identifier"""
        try:
            return (
                self.__invD[entryId.upper()]["content_type"][contentType]
                if isinstance(self.__invD[entryId.upper()]["content_type"][contentType], list)
                else [self.__invD[entryId.upper()]["content_type"][contentType]]
            )
        except Exception as e:
            logger.debug("Failing for %r %r with %s", entryId, contentType, str(e))
        return []

    def getInventory(self):
        """Return the removed inventory dictionary"""
        try:
            return self.__invD
        except Exception as e:
            logger.debug("Failing with %s", str(e))
        return {}

    def getAllContentTypes(self):
        """Return the removed content types for the input entry identifier"""
        entryId = None
        try:
            cS = set()
            for entryId, tD in self.__invD.items():
                if "content_type" in tD:
                    for cT in tD["content_type"].keys():
                        cS.add(cT)
            return sorted(list(cS))
        except Exception as e:
            logger.exception("Failing for %r with %s", entryId, str(e))
        return []

    def __reload(self, urlTarget, urlFallbackTarget, dirPath, useCache=True):
        invD = {}
        fU = FileUtil()
        fn = fU.getFileName(urlTarget)
        fp = os.path.join(dirPath, fn)
        self.__mU.mkdir(dirPath)
        #
        if self.__storeCache and useCache and self.__mU.exists(fp):
            invD = self.__mU.doImport(fp, fmt="json")
            logger.info("Reading cached inventory (%d) from file %s", len(invD), fp)
        else:
            invD = self.__mU.doImport(urlTarget, fmt="json")
            logger.info("Loaded inventory from %s (%r)", urlTarget, len(invD))
            if len(invD) == 0:
                invD = self.__mU.doImport(urlFallbackTarget, fmt="json")
                logger.info("Loaded fallback inventory from %s (%r)", urlFallbackTarget, len(invD))
        #
        if self.__storeCache:
            logger.info("Fetch inventory from %s", urlTarget)
            ok = fU.get(urlTarget, fp)
            if not ok:
                ok = fU.get(urlFallbackTarget, fp)
        #
        return invD

    def getStatusDetails(self, curD):
        rmD = {}
        for entryId, tD in self.__invD.items():
            if entryId in curD:
                continue
            #
            if tD["status_code"] == "TRSF":
                rmD[entryId] = {"status": "REMOVED", "status_code": "TRSF"}
            elif tD["status_code"] == "OBS":
                sL = self.getSupersededBy(entryId)
                if sL:
                    if sL[-1] in curD:
                        rmD[entryId] = {"status": "REMOVED", "status_code": "OBS", "id_code_replaced_by_latest": sL[-1]}
                    else:
                        rmD[entryId] = {"status": "REMOVED", "status_code": "OBS"}
                else:
                    rmD[entryId] = {"status": "REMOVED", "status_code": "OBS"}
        return rmD

    def getRcsbRemovedData(self):
        """Assemble and reorganize removed data for RCSB loading operations.

        Returns:
            dict: dictionary of transferred data,
            dict: dictionary of removed onsilico data,
            dict: dictioary of audit authors of removed data,
            dict: dictionary of all removed data,
            dict: dictionary of superseding entries
        """
        return self.__assembleRemovedData(self.__invD)

    def __assembleRemovedData(self, invD):
        """Assemble and reorganize removed data for RCSB loading operations.

        Args:
            invD (dict): json inventory of removed data

        Returns:
            dict: dictionary of transferred data,
            dict: dictionary of removed onsilico data
            dict: dictioary of audit authors of removed data
            dict: dictionary of all removed data
            dict: dictionary of superseding entries
        """
        # for transferred (which are the theoretical models, only pdb format is available):
        ct1MapD = {
            "pdb": "coordinates",
        }
        # for removed:
        ct2MapD = {
            "combined_nmr_data_nef": "Combined NMR data (NEF)",
            "combined_nmr_data_nmr-star": "Combined NMR data (NMR-STAR)",
            "nmr_chemical_shifts": "NMR chemical shifts",
            "nmr_restraints_v1": "NMR restraints V1",
            "nmr restraints v2": "NMR restraints V1",
            # "assembly PDB",
            # "assembly mmCIF",
            "pdb": "entry PDB",
            # "entry PDB bundle",
            "pdbml": "entry PDBML",
            "mmcif": "entry mmCIF",
            "structure_factors": "structure factors",
        }
        trMapD = {
            "deposition_authors": "audit_authors",
            "status_code": "status_code",
            "release_date": "release_date",
            "remote_repository_name": "remote_repository_name",
            "remote_repository_title": "title",
            "content_type": "repository_content_types",
            "deposit_date": "deposit_date",
            "remote_accession_code": "remote_accession_code",
            # "title": "title",
        }
        insMapD = {
            "deposition_authors": "audit_authors",
            "deposit_date": "deposit_date",
            "superseded_by": "id_codes_replaced_by",
            "release_date": "release_date",
            "obsolete_date": "remove_date",
            "status_code": "status_code",
            "title": "title",
        }
        removedMapD = {
            "deposition_authors": "audit_authors",
            "deposit_date": "deposit_date",
            "details": "details",
            "superseded_by": "id_codes_replaced_by",
            "release_date": "release_date",
            "obsolete_date": "remove_date",
            "content_type": "repository_content_types",
            "title": "title",
        }
        dateFields = ["deposit_date", "release_date", "remove_date"]
        trsfD = {}
        insilicoD = {}
        auditAuthorD = {}
        removedD = {}
        superD = {}
        # --- generate lookup for superseding entries ---
        replacedByD = {}
        replacesD = {}
        for entryId, tD in invD.items():
            if "superseded_by" in tD:
                if isinstance(tD["superseded_by"], str):
                    replacedByD[entryId] = [tD["superseded_by"].upper()]
                else:
                    replacedByD[entryId] = [t.upper() for t in tD["superseded_by"]]
        for entryId, rIdL in replacedByD.items():
            for rId in rIdL:
                replacesD.setdefault(rId, []).append(entryId)
        # ---
        for entryId, tD in invD.items():
            # --- Transferred ---
            if tD["status_code"] == "TRSF":
                qD = {}
                for iky, oky in trMapD.items():
                    if iky not in tD:
                        continue
                    if self.__assignDates and oky in dateFields:
                        qD[oky] = dateutil.parser.parse(tD[iky])
                    elif oky in dateFields:
                        qD[oky] = tD[iky][:10]
                    else:
                        qD[oky] = tD[iky]
                    #
                    if iky == "content_type":
                        cS = set()
                        for ct in tD["content_type"]:
                            if ct in ct1MapD:
                                cS.add(ct1MapD[ct])
                        qD["repository_content_types"] = sorted(list(cS))
                    #
                    if oky == "id_codes_replaced_by":
                        if isinstance(qD["id_codes_replaced_by"], str):
                            qD["id_codes_replaced_by"] = [qD["id_codes_replaced_by"].upper()]
                        else:
                            qD["id_codes_replaced_by"] = [t.upper() for t in qD["id_codes_replaced_by"]]
                trsfD[entryId] = qD
            else:
                # --- removed ---
                qD = {}
                for iky, oky in removedMapD.items():
                    if iky not in tD:
                        continue
                    if self.__assignDates and oky in dateFields:
                        qD[oky] = dateutil.parser.parse(tD[iky])
                    elif oky in dateFields:
                        qD[oky] = tD[iky][:10]
                    else:
                        qD[oky] = tD[iky]
                    #
                    if iky == "content_type":
                        cS = set()
                        for ct in tD["content_type"]:
                            if ct in ct2MapD:
                                cS.add(ct2MapD[ct])
                        qD["repository_content_types"] = sorted(list(cS))
                    #
                    if oky == "id_codes_replaced_by":
                        if isinstance(qD["id_codes_replaced_by"], str):
                            qD["id_codes_replaced_by"] = [qD["id_codes_replaced_by"].upper()]
                        else:
                            qD["id_codes_replaced_by"] = [t.upper() for t in qD["id_codes_replaced_by"]]
                    #
                removedD[entryId] = qD
            #
            # --- inslico models
            if tD["status_code"] in ["OBS", "TRSF", "WDRN"] and self.__isContentInsilico(tD["content_type"]):
                qD = {}
                for iky, oky in insMapD.items():
                    if iky not in tD:
                        continue
                    # yyyy-mm-dd
                    if self.__assignDates and oky in dateFields:
                        qD[oky] = dateutil.parser.parse(tD[iky])
                    elif oky in dateFields:
                        qD[oky] = tD[iky][:10]
                    else:
                        qD[oky] = tD[iky]
                    if oky == "id_codes_replaced_by":
                        if isinstance(qD["id_codes_replaced_by"], str):
                            qD["id_codes_replaced_by"] = [qD["id_codes_replaced_by"].upper()]
                        else:
                            qD["id_codes_replaced_by"] = [t.upper() for t in qD["id_codes_replaced_by"]]
                #
                insilicoD[entryId] = qD
            # --- audit authors ---
            if "deposition_authors" in tD:
                qL = []
                for ii, author in enumerate(tD["deposition_authors"], 1):
                    qL.append({"ordinal_id": ii, "audit_author": author})
                auditAuthorD[entryId] = qL

            #  ---- superseded ----
            if entryId in replacesD:
                superD[entryId] = {"id_codes_superseded": replacesD[entryId]}
        logger.info("# of transferred entries: %d", len(trsfD))
        logger.info("# of insilico entries: %d", len(insilicoD))
        logger.info("# of removed entries: %d", len(removedD))
        return trsfD, insilicoD, auditAuthorD, removedD, superD

    def __isContentInsilico(self, ctD):
        """Test if the content type dictionary contains an inslico model.

        Args:
            ctD (dict): content type dictionary

        Returns:
            (bool): True if content is insilico

        Example:

            "content_type": {
                    "pdb": [
                        "/pdb/data/structures/models/current/pdb/ir/pdb2ir4.ent.gz"
                    ]
                },
        """
        isModel = False
        for ct, pthL in ctD.items():
            if ct in ["pdb", "mmcif", "pdbml"]:
                for pth in pthL:
                    if pth.startswith("/pdb/data/structures/models"):
                        isModel = True
                        break
        return isModel
